@@ -11,10 +11,11 @@
 # limitations under the License.
 
 from torchdyn.core.defunc import DEFunc
-from torchdyn.core.neuralde import NeuralODE, NeuralSDE
+from torchdyn.core.neuralde import NeuralODE, NeuralSDE, MultipleShootingLayer
 from torchdyn.core.problems import ODEProblem, SDEProblem, MultipleShootingProblem
 
 # backward-compatibility (pre v0.2.0)
 NeuralDE = NeuralODE
 
-__all__ =   ['DEFunc', 'NeuralODE', 'NeuralDE', 'NeuralSDE', 'ODEProblem', 'SDEProblem', 'MultipleShootingProblem']
\ No newline at end of file
+__all__ =   ['DEFunc', 'NeuralODE', 'NeuralDE', 'NeuralSDE', 'ODEProblem', 'SDEProblem', 
+            'MultipleShootingProblem', 'MultipleShootingLayer']
\ No newline at end of file
