@@ -1,3 +1,15 @@
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+#      http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
+
 """
     Contains ODE solvers, both sequential as well as time-parallel multiple shooting methods necessary for multiple-shoting layers [1]. 
     The stateful design allows users to modify or tweak each Tableau during training, ensuring compatibility with hybrid methods such as Hypersolvers [2]
@@ -12,15 +24,7 @@ from torchdyn.numerics._constants import construct_rk4, construct_dopri5, constr
 
 
 class SolverTemplate(nn.Module):
-    def __init__(self, order, min_factor=0.2, max_factor=10, safety=0.9):
-        """[summary]
-
-        Args:
-            order ([type]): [description]
-            min_factor (float, optional): [description]. Defaults to 0.2.
-            max_factor ([type], optional): [description]. Defaults to 10..
-            safety (float, optional): [description]. Defaults to 0.9.
-        """
+    def __init__(self, order, min_factor:float=0.2, max_factor:float=10, safety:float=0.9):
         super().__init__()
         self.order = order
         self.min_factor = torch.tensor([min_factor])
@@ -46,6 +50,7 @@ class SolverTemplate(nn.Module):
 
 class Euler(SolverTemplate):
     def __init__(self, dtype=torch.float32):
+        """Explicit Euler ODE stepper, order 1"""
         super().__init__(order=1)
         self.dtype = dtype
         self.stepping_class = 'fixed'
@@ -58,6 +63,7 @@ class Euler(SolverTemplate):
     
 class Midpoint(SolverTemplate):
     def __init__(self, dtype=torch.float32):
+        """Explicit Midpoint ODE stepper, order 2"""
         super().__init__(order=2)
         self.dtype = dtype
         self.stepping_class = 'fixed'
@@ -71,6 +77,7 @@ class Midpoint(SolverTemplate):
     
 class RungeKutta4(SolverTemplate):
     def __init__(self, dtype=torch.float32):
+        """Explicit Midpoint ODE stepper, order 4"""
         super().__init__(order=4)
         self.dtype = dtype
         self.stepping_class = 'fixed'
@@ -88,6 +95,8 @@ class RungeKutta4(SolverTemplate):
 
 class AsynchronousLeapfrog(SolverTemplate):
     def __init__(self, channel_index:int=-1, stepping_class:str='fixed', dtype=torch.float32):
+        """Explicit Leapfrog symplectic ODE stepper. 
+        Can return local error estimates if adaptive stepping is required"""
         super().__init__(order=2)
         self.dtype = dtype
         self.channel_index = channel_index
@@ -117,12 +126,7 @@ class AsynchronousLeapfrog(SolverTemplate):
 
 class DormandPrince45(SolverTemplate):
     def __init__(self, dtype=torch.float32):
-        """[summary]
-
-        Args:
-            dtype ([type], optional): [description]. Defaults to torch.float32.
-        """
-        super().__init__(order=6)
+        super().__init__(order=5)
         self.dtype = dtype
         self.stepping_class = 'adaptive'
         self.tableau = construct_dopri5(self.dtype)
@@ -138,13 +142,13 @@ class DormandPrince45(SolverTemplate):
         k7 = f(t + c[5] * dt, x + dt * a[5][0] * k1 + dt * a[5][1] * k2 + dt * a[5][2] * k3 + dt * a[5][3] * k4 + dt * a[5][4] * k5 + dt * a[5][5] * k6)
         x_sol = x + dt * (bsol[0] * k1 + bsol[1] * k2 + bsol[2] * k3 + bsol[3] * k4 + bsol[4] * k5 + bsol[5] * k6)
         err = berr[0] * k1 + berr[1] * k2 + berr[2] * k3 + berr[3] * k4 + berr[4] * k5 + berr[5] * k6 + berr[6] * k7
-        return k7, x_sol, err, (k1, k2, k3, k4 ,k5, k6, k7)
+        return k7, x_sol, err, (k1, k2, k3, k4, k5, k6, k7)
 
 
 
 class Tsitouras45(SolverTemplate):
     def __init__(self, dtype=torch.float32):
-        super().__init__(order=6)
+        super().__init__(order=5)
         self.dtype = dtype
         self.stepping_class = 'adaptive'
         self.tableau = construct_tsit5(self.dtype)
@@ -160,7 +164,7 @@ class Tsitouras45(SolverTemplate):
         k7 = f(t + c[5] * dt, x + dt * a[5][0] * k1 + dt * a[5][1] * k2 + dt * a[5][2] * k3 + dt * a[5][3] * k4 + dt * a[5][4] * k5 + dt * a[5][5] * k6)
         x_sol = x + dt * (bsol[0] * k1 + bsol[1] * k2 + bsol[2] * k3 + bsol[3] * k4 + bsol[4] * k5 + bsol[5] * k6)
         err = berr[0] * k1 + berr[1] * k2 + berr[2] * k3 + berr[3] * k4 + berr[4] * k5 + berr[5] * k6 + berr[6] * k7
-        return k7, x_sol, err, (k1, k2, k3, k4 ,k5, k6, k7)
+        return k7, x_sol, err, (k1, k2, k3, k4, k5, k6, k7)
 
 
 class ImplicitEuler(SolverTemplate):
@@ -222,8 +226,8 @@ class MSZero(MShootingSolverTemplate):
         """Multiple shooting solver using Parareal updates (zero-order approximation of the Jacobian)
 
         Args:
-            coarse_method (str, optional): [description]. Defaults to 'euler'.
-            fine_method (str, optional): [description]. Defaults to 'rk4'.
+            coarse_method (str, optional): . Defaults to 'euler'.
+            fine_method (str, optional): . Defaults to 'rk4'.
         """
         super().__init__(coarse_method, fine_method)
 
@@ -252,8 +256,8 @@ class MSBackward(MShootingSolverTemplate):
         """Multiple shooting solver using discrete adjoints for the Jacobian
 
         Args:
-            coarse_method (str, optional): [description]. Defaults to 'euler'.
-            fine_method (str, optional): [description]. Defaults to 'rk4'.
+            coarse_method (str, optional): . Defaults to 'euler'.
+            fine_method (str, optional): . Defaults to 'rk4'.
         """
         super().__init__(coarse_method, fine_method)
 
@@ -263,7 +267,6 @@ class MSBackward(MShootingSolverTemplate):
         i = 0
         B = B.requires_grad_(True)
         while i <= maxiter:
-            print(i)
             i += 1
             B_fine = odeint_func(f, B[i-1:], sub_t_span, solver=self.fine_method)[1][-1]
             B_out = torch.zeros_like(B)
@@ -282,14 +285,12 @@ class MSBackward(MShootingSolverTemplate):
 
 class ParallelImplicitEuler(MShootingSolverTemplate):
     def __init__(self, coarse_method='euler', fine_method='euler'):
-        """Parallel Implicit Euler Method
-        """
+        """Parallel Implicit Euler Method"""
         super().__init__(coarse_method, fine_method)
         self.solver = torch.optim.LBFGS
         self.max_iters = 200
 
     def sync_device_dtype(self, x, t_span):
-        "Ensures `x`, `t_span`, `tableau` and other solver tensors are on the same device with compatible dtypes"
         return x, t_span
 
     @staticmethod
